/*
  Tiny simple standalone implementation of Elliptic Curve Cryptography, 
  using a single hardcoded NIST curve.

  Recommended only for educational use, since it may have unknown holes.
  Definitely slow, and subject to side channel (time, heat, power) attacks.
  
  Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2013-03-25 (Public Domain)
*/
#include <iostream>
#include "bigint.h"
#include "bigint.cpp" /* for linking easier */

int main(int argc,char *argv[]) {
	// All curve points are modulo this huge prime:
	BigInteger p; p.readHex("FFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF");
	
	// The elliptic curve here is NIST p256r1: y^2 = x^3 - 3*x + b
	//    b is thus the constant part of the curve.
	BigInteger b; b.readHex("5AC635D8AA3A93E7B3EBBD55769886BC651D06B0CC53B0F63BCE3C3E27D2604B");
	
	// Here's a NIST starting point on the curve.
	BigInteger x; x.readHex("6B17D1F2E12C4247F8BCE6E563A440F277037D812DEB33A0F4A13945D898C296");
	BigInteger y; y.readHex("4FE342E2FE1A7F9B8EE7EB4A7C0F9E162BCE33576B315ECECBB6406837BF51F5");
	
	//  Verify that the starting point lies on the curve.
	std::cout<<"p="<<p<<"\n";
	BigInteger nonmod=x*x*x-3*x+b-y*y;
	std::cout<<"x,y="<<x<<" , "<<y<<"\n";
	std::cout<<"Non-modular curve value: "<<nonmod<<"\n";
	std::cout<<"Modulo p: "<<nonmod.mod(p)<<"\n";

for (int dub=0;dub<10;dub++) {
	std::cout<<"------------- doubling step "<<dub<<" -------------\n";
	// Compute local tangent, and use it to double the starting point.
	BigInteger m=(3*x*x - 3)*(2*y).modInverse(p); // == dy/dx for curve
	BigInteger v = y-m*x; // intercept from point on curve
	
	/*
	  Algebra interlude:
x^3 - 3 x + b = y^2
= (m x + v)^2
= m^2 x^2 + 2 m v x + v^2

So
  0 = x^3 - m^2 x^2 - (2 m v + 3) + (b-v^2)
Since it's a cubic, it's got three roots.  Write them as
  0 = (x-p1.x)*(x-p2.x)*(x-p3.x)
By matching the x^2 term, we get:
  -p1.x-p2.x-p3.x = -m^2
So
  p1.x+p2.x+p3.x = m^2
Thus
  p3.x = m^2 - p1.x - p2.x
  p3.y = m*p3.x+v
	*/
	BigInteger x2 = m*m - 2*x; // x from matching x^2 terms
	BigInteger y2 = -(m*x2+v); // y from slope equation, plus mirroring

	if (dub<1) { // do the non-modular version *once* (it's slow!)
		// This should be a new curve point--check it.
		std::cout<<"x2,y2="<<x2<<" , "<<y2<<"\n";
		BigInteger nonmod2=x2*x2*x2-3*x2+b-y2*y2;
		std::cout<<"Non-modular value: "<<nonmod2<<"\n";
		std::cout<<"Modulo p: "<<nonmod2.mod(p)<<"\n";
	}

	// mod-p versions of x2 and y2 should still be curve points.  Verify.
	x2=x2.mod(p);
	y2=y2.mod(p);
	std::cout<<"Mod(p) x2,y2="<<x2<<" , "<<y2<<"\n";
	BigInteger nonmod2=((x2*x2).mod(p)*x2).mod(p)-3*x2+b-(y2*y2).mod(p);
	//std::cout<<"Non-modular value: "<<nonmod2<<"\n";
	std::cout<<"Modulo p: "<<nonmod2.mod(p)<<"\n";
	
	x=x2;
	y=y2;
} // go around doubling loop again
	
	return 0;
}

