/*
  A toy implementation of the Digital Signature Algorithm (DSA)
  using Elliptic Curve Cryptography (ECC).

  CAUTION -- QUITE INSECURE: 
    - Vulnerable to side-channel attacks, like time, heat, and power.
    - Uses srandom/random to generate k and secret key (fail)
  Recommended only for classroom use.
  
  This is implemented directly from Wikipedia (but it works!)
     https://en.wikipedia.org/wiki/Elliptic_Curve_DSA
  
  Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2013-03-29 (Public Domain)
*/
#include <iostream>
#include "ecc_lib.h"
#include "ecc_lib.cpp" /* for linking easier */


int main(int argc,char *argv[]) {
	if (argc>1) srandom(atoi(argv[1])); // <- FAIL.  Only 32 bits of randomness!
	
	int bits=256; // number of bits in each scalar integer
	if (argc>2) bits=atoi(argv[2]);

	// Pick a shared elliptic curve
	//ECcurve_NISTP256 curve; // NIST P-256
	//ECcurve_BitCoin curve; // SECG p256k1
	ECcurve_brainpoolP512r1 curve; // ecc brainpool

	// Pick a private key, and derive the public key
	BigInteger privateKey=BigInteger::random(bits); // dA
	ECpoint publicKey=curve.start.multiply(privateKey,curve); // QA
	std::cout<<"Public key: "<<publicKey<<"\n";
	
	// You can check a public key is valid by multiplying by n.
	//   This should hit infinity.
	ECpoint check=publicKey.multiply(curve.n,curve);
	std::cout<<"Public key order check: "<<check<<"\n";

	// The hash should be computed from the message (not random!)
	BigInteger z=BigInteger::random(bits);
	
// Signature generation:
	std::cout<<"Hash: ( "<<z.hex()<<" )\n";
	BigInteger r,s; // These are the signature
	do {
	
		// Each message requires a random, unique k
		//   CAUTION: reusing a k even once reveals the key!
		BigInteger k=1+BigInteger::random(bits);
		ECpoint kG=curve.start.multiply(k,curve); 
	
		// r is the first part of the signature
		//   It's the x coordinate of k*G, mod the curve order n
		r=kG.x.mod(curve.n);
		std::cout<<" r= "<<r<<"\n";
	
		// s is the second part of the signature
		//   It's k inverse * (hash + r*privateKey)
		BigInteger kInverse=k.modInverse(curve.n);
		s=(kInverse*(z+r*privateKey)).mod(curve.n);
		std::cout<<" s= "<<s<<"\n";
	} while (r==0 || s==0);
	
	
// Signature verification
	// FIXME: verify that r and s are >=1 and <curve.n
	BigInteger w=s.modInverse(curve.n);
	BigInteger u1=(z*w).mod(curve.n);
	BigInteger u2=(r*w).mod(curve.n);
	ECpoint R=
	    curve.start.multiply(u1,curve)
	  .add(
	    publicKey.multiply(u2,curve),
	  curve);
	
	if (r==R.x.mod(curve.n)) std::cout<<"Signature valid!  :-)\n";
	else {
		std::cout<<"Signature not valid.  :-(\n";
		std::cout<<"             r= "<<r<<"\n";
		std::cout<<" but u1*G+u2*K= "<<R.x.mod(curve.n)<<"\n";
	}	
	return 0;
}

