/**
  Elliptic curve example, structured as a library.
  
  Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2013-03-25 (Public Domain)
*/
#include "ecc_lib.h"
#include "bigint.cpp" /* library body, for easier linking. */

/* The elliptic curve point "infinity" is the additive identity (0) */
ECpoint ECpoint::infinity(-1,ECcoord(2).pow(10000)); // x==-1, y==big

ECpoint ECpoint::add(const ECpoint &o,const ECcurve &curve) const
{
	// special cases for infinity:
	if (*this == infinity) return o; /* we're zero */
	if (o == infinity) return *this; /* they're zero */

	// Usual case:
	const ECcoord &p=curve.p;
	ECcoord m;
	if (x-o.x==0) { // special case for same X coordinate
		if (y-o.y==0) { // adding point to itself--take tangent
			m=curve.tangent(x,y);
		} else { // adding additive inverses--return "infinity"
			return infinity;
		}
	}
	else { // default case: 
		m=(y-o.y) * (x-o.x).modInverse(p); // dy/dx slope of line (default case)
	}
	m=m.mod(p); // modular arithmetic, to speed things up here
	ECcoord v = y-m*x; // line's y intercept

	ECcoord x3 = (m*m - x - o.x).mod(p); // comes from matching x^2 terms
	ECcoord y3 = (-(m*x3+v)).mod(p); // from line equation, plus mirroring

	return ECpoint(x3,y3);
}

// Multiply us by this scalar, under the action of this curve.
ECpoint ECpoint::multiply(BigInteger scalar,const ECcurve &curve) const
{
	ECpoint sum=ECpoint::infinity; 
	ECpoint A=*this;
	for (int bit=0;scalar!=0;bit++) {
		if ( bi_is_odd(scalar.get()) ) // this bit is set--include value in sum
		{ // do sum+=A;
			sum=sum.add(A,curve);
		}
		scalar = scalar/2; // shift right by 1 bit
		if (scalar!=0) // more data is available
			A=A.add(A,curve); // shift A up to next bit (A+=A)
		// possible optimization: save the table of A values...
	}
	return sum;
}


#ifdef STANDALONE /* Simple standalone test program */

#include "osl/time_function.h"

// Timing code for point multiplication by scalar
int time_scalar=0; // value we're multiplying by
ECpoint time_fearme; // don't let compiler's optimizer skip the multiply
int time_scalar_mul(void) {
	static ECcurve_brainpoolP512r1 curve;
	time_fearme=curve.start.multiply(time_scalar,curve);
	return 0;
}

int main(int argc,char *argv[]) {
	int last_i=1000;
	if (argc>1) last_i=atoi(argv[1]);

/* Check timing attack */
	time_scalar=0; print_time("time_scalar(0)",time_scalar_mul);
	time_scalar=0xffffff; print_time("time_scalar(0xffffff)",time_scalar_mul);

/* Check correctness of scalar multiplication via test vector */
{
	// One test vector, from an OpenSSL-generated ECDSA private key
	ECcurve_secp521r1 curve;
	
	ECcoord priv; priv.readHex(
    "19:5c:95:f3:02:07:bb:b1:ea:48:16:1f:11:2f:1b:"
    "c7:89:d9:ae:d2:b3:a5:80:51:92:3f:9e:d2:bc:0b:"
    "17:12:75:8c:22:7f:50:3d:7c:77:e6:3d:07:75:02:"
    "4b:f9:8c:31:fc:41:fa:d4:65:df:52:2c:49:42:a3:"
    "a0:b4:52:4f:18");
	ECpoint pub=curve.start.multiply(priv,curve);
	/*
	std::cout<<"ECDSA private key="<<priv.hex()<<"\n";
	std::cout<<"ECDSA public key="<<pub.x.hex()<<" , "<<pub.y.hex()<<"\n";
	*/
	ECcoord true_x; true_x.readHex("e0cfe3990ac506e4bfaf9692bdaf2c8d53d01896dbfd7e691828bde469f027dc9b534a2a856214582f6c5ed76f497942b9caf0b84d0203b7f54c82edcbaa83a5f5");
	if (pub.x!=true_x)
	{
		std::cout<<"EEK!  Scalar multiplcation returned incorrect value!\n";
		std::cout<<"  We should have gotten "<<true_x<<", but we got "<<pub.x<<"\n";
		abort();
	}
}

/* Move some points along the curve */
	//ECcurve_NISTP256 curve; // NIST P-256
	//ECcurve_BitCoin curve; // SECG p256k1
	ECcurve_brainpoolP512r1 curve;
	ECpoint c=curve.start;
	
	// Verify that multiplication-by-scalar actually works
	for (int verify=0;verify<20;verify++) {
		srand(verify);
		int s0=rand()%0xf, s1=rand()%0x3f, s2=rand()%0x7f;
		ECpoint c0=c.multiply(s0,curve);

		ECpoint c1=c0.multiply(s1,curve);
		ECpoint c2=c0.multiply(s2,curve);
		ECpoint c12=c1.add(c2,curve); // c0*s1 + c0*s2

		ECpoint c1p2=c0.multiply(s1+s2,curve); // c0*(s1+s2)
		if (!(c12 == c1p2)) 
		{
			std::cout<<"EEK!  Curve 'multiply' doesn't distribute over addition!\n";
			std::cout<<" verify="<<verify<<"\n";
			std::cout<<" scalars: "<<s0<<" and "<<s1<<" and "<<s2<<"\n";
			std::cout<<" curve points: "<<c12<<" vs "<<c1p2<<"\n";
			abort();
		}
	}
	
	// Print some test multiplications
	for (int i=0;i<last_i;i++) 
	{
		ECpoint ci=c.multiply(i,curve);
		if (false && (i%100 == 1)) 
		{ // periodically print location
			std::cout<<"c*"<<i<<" = "<<ci<<"\n";
		}
		
		if (i>0) { // verify we're still on the curve
			ECcoord z=curve.evaluate(ci.x,ci.y).mod(curve.p);
			if (z!=0) {
				std::cout<<"EEK!  Fell off curve at "<<ci<<" (z="<<z<<", i="<<i<<")\n";
				exit(1);
			}
		}
	}
	
	std::cout<<"Elliptic curve functions pass self-tests properly.\n";
	
	return 0;
}

#endif
