/**
  Trivial Elliptic Curve Cryptography (ECC) code, structured as a library.
 
  Nobody should use this for real secrets, since it's (1) slow, 
  (2) unproven, and (3) vulnerable to timing attacks.
   
  Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2013-03-25 (Public Domain)
*/
#ifndef __OSL_ECC_LIB_H
#define __OSL_ECC_LIB_H

#include "bigint.h"

// Represents the X or Y coordinate of an elliptic curve point.
//   The math only requires a "field" here; typically you use
//   either Z_p (modulo a big prime) or the binary field GF(2^m).
//   I only support the big prime so far.
typedef BigInteger ECcoord;

// Rational number type (used to avoid modInverse in scalar multiplication)
typedef BigRat ECrat;

// Forward declaration of an elliptic curve
class ECcurve;


// One elliptic curve point, represented by 2D coordinates (x,y).
class ECpoint {
public:
   ECcoord x,y; // coordinates of point on the curve
   
   // Normal initialization
   ECpoint(const ECcoord &x_,const ECcoord &y_) :x(x_), y(y_) {}
   
   // By default, points are initialized to infinity.
   ECpoint() :x(infinity.x), y(infinity.y) {}
   
   // This is the "point at infinity".  It's arbitrarily far up.
   //   In the group of ECpoints, it's the identity element.
   static ECpoint infinity; 
   
   friend std::ostream &operator<<(std::ostream &o,const ECpoint &p) {
   	if (p == ECpoint::infinity) o<<"[infinity]";
   	else o<<p.x<<" , "<<p.y;
	return o;
   }
   
   bool operator==(const ECpoint &p) const {
      return x==p.x && y==p.y;
   }
   bool operator!=(const ECpoint &p) const {
      return !operator==(p);
   }
   
 
   // Add us to o, under the action of this curve.
   ECpoint add(const ECpoint &o,const ECcurve &curve) const;
   
   // Multiply us by this scalar, under the action of this curve.
   ECpoint multiply(BigInteger scalar,const ECcurve &curve) const;
};

/******************* curves *************************/
// Abstract superclass of modulo-a-prime type elliptic curves 
//  I need to extend this to support binary GF(2^m) fields.
class ECcurve {
public:
   // All curve arithmetic is modulo this huge prime:
    BigInteger p;

   // Here's an arbitrary starting point on the curve.
   //  This is called a curve generator G.
    ECpoint start;

   // This is the order of the start point
    BigInteger n;

    // Evaluate the elliptic curve at this point. 
    //  Returns 0 if the point lies along our curve.
    virtual ECcoord evaluate(const ECcoord &x,const ECcoord &y) const =0;

    // Evaluate the slope of the tangent of our curve at this curve point.
    virtual ECcoord tangent(const ECcoord &x,const ECcoord &y) const =0;
    
	// Tangent of curve, rational version
	virtual ECrat tangentRat(const ECrat &x,const ECrat &y) const =0;
};

// The elliptic curve here is NIST p256r1: y^2 = x^3 - 3*x + b
//   http://csrc.nist.gov/groups/ST/toolkit/documents/dss/NISTReCur.pdf
//   See page 8, curve P-256.
class ECcurve_NISTP256 : public ECcurve {
public:
    //    b is thus the constant part of the curve.
    ECcoord b;
   
    // Constructor sets initial values.
    ECcurve_NISTP256() {
	p.readHex("FFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF");
	n=BigInteger("115792089210356248762697446949407573529996955224135760342422259061068512044369");
	b.readHex("5AC635D8AA3A93E7B3EBBD55769886BC651D06B0CC53B0F63BCE3C3E27D2604B");

       start.x.readHex("6B17D1F2E12C4247F8BCE6E563A440F277037D812DEB33A0F4A13945D898C296");
       start.y.readHex("4FE342E2FE1A7F9B8EE7EB4A7C0F9E162BCE33576B315ECECBB6406837BF51F5");
    }

    // Evaluate the elliptic curve at this point. 
    //  Returns 0 if the point lies along our curve.
    inline ECcoord evaluate(const ECcoord &x,const ECcoord &y) const {
         return ((x*x).mod(p)*x-3*x+b-y*y).mod(p);   // 0 = x^3 - 3*x + b - y^2
    }

    // Evaluate the slope of the tangent of our curve at this curve point.
    inline ECcoord tangent(const ECcoord &x,const ECcoord &y) const {
         return ((3*x*x).mod(p) - 3)*(2*y).modInverse(p); // == dy/dx
    }
	// Tangent of curve, rational version
	inline ECrat tangentRat(const ECrat &x,const ECrat &y) const {
		return (3*x*x-3).mod(p)/(2*y); // == dy/dx
	}
};


// The elliptic curve here is SECG secp256k1: y^2 = x^3 + 7
//   See parameters in http://www.secg.org/collateral/sec2_final.pdf
//   section 2.7.1.  It's used by the BitCoin protocol.
class ECcurve_BitCoin : public ECcurve {
public:
   
    // Constructor sets initial values.
    ECcurve_BitCoin() {
	p.readHex("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F");
	n.readHex("FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFE BAAEDCE6 AF48A03B BFD25E8C D0364141");

       // SEC's "04" means they're representing the generator point's X,Y parts explicitly.
       //  The compressed "02" form means storing only x (you compute Y)
	start.x.readHex("79BE667E F9DCBBAC 55A06295 CE870B07 029BFCDB 2DCE28D9 59F2815B 16F81798");
	start.y.readHex("483ADA77 26A3C465 5DA4FBFC 0E1108A8 FD17B448 A6855419 9C47D08F FB10D4B8");
    }

    // Evaluate the elliptic curve at this point. 
    //  Returns 0 if the point lies along our curve.
    inline ECcoord evaluate(const ECcoord &x,const ECcoord &y) const {
         return ((x*x).mod(p)*x+7-y*y).mod(p);   // 0 = x^3 +7 - y^2
    }

    // Evaluate the slope of the tangent of our curve at this curve point.
    inline ECcoord tangent(const ECcoord &x,const ECcoord &y) const {
         return (3*x*x).mod(p)*(2*y).modInverse(p); // == dy/dx
    }
	// Tangent of curve, rational version
	inline ECrat tangentRat(const ECrat &x,const ECrat &y) const {
		return (3*x*x).mod(p)/(2*y); // == dy/dx
	}
};

// General form for all brainpool or secg r curves: y^2 = x^3 + Ax + B mod p
class ECcurve_AB : public ECcurve {
public:
	ECcoord A,B;
	
	inline ECcoord evaluate(const ECcoord &x,const ECcoord &y) const {
		return ((x*x).mod(p)*x+A*x+B-y*y).mod(p);   // 0 = x^3 +Ax+B - y^2
	}

	// Evaluate the slope of the tangent of our curve at this curve point.
	inline ECcoord tangent(const ECcoord &x,const ECcoord &y) const {
		return (3*x*x+A).mod(p)*(2*y).modInverse(p); // == dy/dx
	}
	
	// Tangent of curve, rational version
	inline ECrat tangentRat(const ECrat &x,const ECrat &y) const {
		return (3*x*x+A).mod(p)/(2*y); // == dy/dx
	}
};

// The elliptic curve here is SECG secp521r1: y^2 = x^3 + 7
//   See parameters in http://www.secg.org/collateral/sec2_final.pdf
class ECcurve_secp521r1 : public ECcurve_AB {
public: 
    // Constructor sets initial values.
    ECcurve_secp521r1() {
	p.readHex("01FF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF"
"FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF"
"FFFFFFFF FFFFFFFF FFFFFFFF");
	n.readHex("01FF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF"
"FFFFFFFF FFFFFFFA 51868783 BF2F966B 7FCC0148 F709A5D0 3BB5C9B8"
"899C47AE BB6FB71E 91386409");

	A.readHex("01FF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF"
"FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF"
"FFFFFFFF FFFFFFFF FFFFFFFC");
	B.readHex("0051 953EB961 8E1C9A1F 929A21A0 B68540EE A2DA725B 99B315F3"
"B8B48991 8EF109E1 56193951 EC7E937B 1652C0BD 3BB1BF07 3573DF88"
"3D2C34F1 EF451FD4 6B503F00");

       // SEC's "04" means they're representing the generator point's X,Y parts explicitly.
       //  The compressed "02" form means storing only x (you compute Y)
	start.x.readHex("C6858E 06B70404 E9CD9E3E CB662395 B4429C64 8139053F"
"B521F828 AF606B4D 3DBAA14B 5E77EFE7 5928FE1D C127A2FF A8DE3348"
"B3C1856A 429BF97E 7E31C2E5 BD66");
	start.y.readHex("0118 39296A78 9A3BC004 5C8A5FB4"
"2C7D1BD9 98F54449 579B4468 17AFBD17 273E662C 97EE7299 5EF42640"
"C550B901 3FAD0761 353C7086 A272C240 88BE9476 9FD16650");
    }
};

// Brainpool 512-bit curve: 
class ECcurve_brainpoolP512r1 : public ECcurve_AB {
public:
	ECcurve_brainpoolP512r1() {
		p.readHex("AADD9DB8DBE9C48B3FD4E6AE33C9FC07CB308DB3B3C9D20ED6639CCA703308717D4D9B009BC66842AECDA12AE6A380E62881FF2F2D82C68528AA6056583A48F3");
		n.readHex("AADD9DB8DBE9C48B3FD4E6AE33C9FC07CB308DB3B3C9D20ED6639CCA70330870553E5C414CA92619418661197FAC10471DB1D381085DDADDB58796829CA90069");
		start.x.readHex("81AEE4BDD82ED9645A21322E9C4C6A9385ED9F70B5D916C1B43B62EEF4D0098EFF3B1F78E2D0D48D50D1687B93B97D5F7C6D5047406A5E688B352209BCB9F822");
		start.y.readHex("7DDE385D566332ECC0EABFA9CF7822FDF209F70024A57B1AA000C55B881F8111B2DCDE494A5F485E5BCA4BD88A2763AED1CA2B2FA8F0540678CD1E0F3AD80892");
		
		A.readHex("7830A3318B603B89E2327145AC234CC594CBDD8D3DF91610A83441CAEA9863BC2DED5D5AA8253AA10A2EF1C98B9AC8B57F1117A72BF2C7B9E7C1AC4D77FC94CA");
		B.readHex ("3DF91610A83441CAEA9863BC2DED5D5AA8253AA10A2EF1C98B9AC8B57F1117A72BF2C7B9E7C1AC4D77FC94CADC083E67984050B75EBAE5DD2809BD638016F723");	
	}
};

#endif
